import { ConnectionPositionPair } from "@angular/cdk/overlay";
import { ComponentPortal } from "@angular/cdk/portal";
import { Directive, EventEmitter, Injector, Input, Output } from "@angular/core";
import { Subject, fromEvent, merge, EMPTY, of, animationFrameScheduler, timer, interval } from 'rxjs';
import { takeUntil, filter, tap, map, debounce, switchMap, delay } from 'rxjs/operators';
import { TOOLTIP_CONTEXT_TOKEN } from "./config/tooltip-context-token";
import { TooltipContainerComponent } from "./tooltip-container.component";
import { placementFlipMap } from "./config/placement-flip-map";
import * as i0 from "@angular/core";
import * as i1 from "@angular/cdk/overlay";
export class TooltipDirective {
    constructor(_elRef, _overlay, _vcRef, _injector) {
        this._elRef = _elRef;
        this._overlay = _overlay;
        this._vcRef = _vcRef;
        this._injector = _injector;
        /**
         * When true, the tooltip will not not respond to any open or close events. Nor will it
         * respond to changes to the `isOpen` input
         */
        this.isDisabled = false;
        /**
         * Change the placement of the tooltip to its opposite position when it moves outside the viewport
         */
        this.autoFlip = true;
        /**
         * Display an arrow or not. The location of the arrow is dependant on the current `placement`
         */
        this.withArrow = true;
        /**
         * Display a close button or not
         */
        this.withClose = false;
        this.closeOnOutsideClick = false;
        /**
         * WARNING: Use with caution - there are potential performance issues with this
         *
         * Update the position of the tooltip before the next browser repaint. An example of where this may be required is if
         * the tooltip is attached (and open) to an element that transitions or animates to a new position
         */
        this.updatePositionOnAnimationFrame = false;
        /**
         * The screen size at which the tooltip should treated as a popover as there is no hover events on mobile
         */
        this.breakpoint = 767;
        this.nwShown = new EventEmitter();
        this.nwHidden = new EventEmitter();
        this.nwClose = new EventEmitter();
        this._destroyed$ = new Subject();
        this._tooltipArrowSize = 5;
        this._manualToggleEvent$ = new Subject();
        this._cancelDelayedOpen$ = new Subject();
        /**
         * A subject that emits when the TooltipContainerComponent is destroyed
         */
        this._tooltipContainerDestroyed$ = new Subject();
    }
    ngOnInit() {
        this._setInputDefaults();
        this._createOverlay();
        this._subscribeToEvents();
        if (this.isOpen) {
            this._manualToggleEvent$.next(this.isOpen);
        }
        if (this.updatePositionOnAnimationFrame) {
            interval(0, animationFrameScheduler)
                .pipe(filter(_ => this.updatePositionOnAnimationFrame && this._overlayRef.hasAttached()), takeUntil(this._destroyed$))
                .subscribe(() => {
                this._overlayRef.updatePosition();
            });
        }
    }
    ngOnChanges(c) {
        var _a, _b, _c, _d;
        const shouldUpdatePositionStrategy = ['placement', 'withArrow', 'autoFlip']
            .some(inputProp => { var _a, _b, _c; return ((_a = c[inputProp]) === null || _a === void 0 ? void 0 : _a.previousValue) !== ((_b = c[inputProp]) === null || _b === void 0 ? void 0 : _b.currentValue) && !((_c = c[inputProp]) === null || _c === void 0 ? void 0 : _c.firstChange); });
        const isOpenChange = ((_a = c.isOpen) === null || _a === void 0 ? void 0 : _a.previousValue) !== ((_b = c.isOpen) === null || _b === void 0 ? void 0 : _b.currentValue) && !c.isOpen.firstChange;
        const shouldUpdateScrollStrategy = ((_c = c.closeOnScroll) === null || _c === void 0 ? void 0 : _c.previousValue) !== ((_d = c.closeOnScroll) === null || _d === void 0 ? void 0 : _d.currentValue) && !c.closeOnScroll.firstChange;
        if (shouldUpdatePositionStrategy) {
            this._updatePositionStrategy(this.placement);
        }
        if (shouldUpdateScrollStrategy) {
            this._updateScrollStrategy(this.closeOnScroll);
        }
        if (isOpenChange) {
            this._manualToggleEvent$.next(this.isOpen);
        }
    }
    /**
     * Can be called manually from the exported directive to open the tooltip
     */
    show() {
        this._manualToggleEvent$.next(true);
    }
    /**
     * Can be called manually from the exported directive to close the tooltip
     */
    hide() {
        this._manualToggleEvent$.next(false);
    }
    /**
     * Can be called manually from the exported directive to toggle the tooltip
     */
    toggle() {
        const isOpen = this._overlayRef.hasAttached();
        this._manualToggleEvent$.next(!isOpen);
    }
    /**
     * Based on the selector used, choose different default if inputs are not defined
     */
    _setInputDefaults() {
        const getDefaultValue = (currVal, defaultVal) => {
            return currVal !== null && currVal !== void 0 ? currVal : defaultVal;
        };
        const isMobileScreenSize = window.innerWidth < this.breakpoint;
        /**
         * Check for undefined and null, not empty strings. This prevents errors when an empty
         * string is passed in the popover or tooltip input
         */
        if ((this.popover !== undefined && this.popover !== null) || isMobileScreenSize) {
            this.delay = getDefaultValue(this.delay, 0);
            this.openEvents = getDefaultValue(this.openEvents, ["click"]);
            this.closeEvents = getDefaultValue(this.closeEvents, ["click"]);
            this.closeOnScroll = getDefaultValue(this.closeOnScroll, false);
            this.pointerEvents = getDefaultValue(this.pointerEvents, 'auto');
            if (isMobileScreenSize) {
                this.closeOnOutsideClick = true;
            }
        }
        else {
            this.delay = getDefaultValue(this.delay, 500);
            this.openEvents = getDefaultValue(this.openEvents, ["mouseenter"]);
            this.closeEvents = getDefaultValue(this.closeEvents, ["click", "mouseleave"]);
            this.closeOnScroll = getDefaultValue(this.closeOnScroll, true);
            this.pointerEvents = getDefaultValue(this.pointerEvents, 'none');
        }
    }
    _getTooltipContent() {
        return this.tooltip || this.popover;
    }
    _open() {
        if (!this._overlayRef.hasAttached()) {
            const portal = new ComponentPortal(TooltipContainerComponent, this._vcRef, this._createInjector());
            const ref = this._overlayRef.attach(portal);
            this.nwShown.emit();
            return ref;
        }
    }
    _close() {
        if (this._overlayRef.hasAttached()) {
            this._overlayRef.detach();
            this.nwHidden.emit();
        }
    }
    _createOverlay() {
        const positionStrategy = this._getPositionStrategy(this.placement);
        const scrollStrategy = this._getScrollStrategy(this.closeOnScroll);
        const disposeOnNavigation = true;
        const panelClasses = ["tooltip-overlay", `pointer-events-${this.pointerEvents}`];
        this._overlayRef = this._overlay.create({ positionStrategy, scrollStrategy, disposeOnNavigation, panelClass: panelClasses });
        if (this.hostElementZIndex) {
            this._overlayRef.hostElement.style.zIndex = this.hostElementZIndex.toString();
        }
    }
    /**
     * Create and return a custom injector that provides the tooltip text as an injectable dependency
     */
    _createInjector() {
        const tooltipData = {
            tooltip: this._getTooltipContent(),
            containerClass: this.containerClass,
            withArrow: this.withArrow,
            withClose: this.withClose,
            templateRefContext: this.context
        };
        return Injector.create({
            parent: this._injector,
            providers: [
                { provide: TOOLTIP_CONTEXT_TOKEN, useValue: tooltipData }
            ]
        });
    }
    _subscribeToEvents() {
        const openEvents$ = this.openEvents.map(eventName => {
            return fromEvent(this._elRef.nativeElement, eventName).pipe(filter(_ => !this._overlayRef.hasAttached()), map(_ => true));
        });
        const closeEvents$ = this.closeEvents.map(eventName => {
            return fromEvent(this._elRef.nativeElement, eventName).pipe(tap(_ => this._cancelDelayedOpen$.next()), filter(_ => this._overlayRef.hasAttached()), map(_ => false));
        });
        const outsideClick$ = this.closeOnOutsideClick ?
            this._overlayRef.outsidePointerEvents()
                .pipe(filter(_ => this._overlayRef.hasAttached()), 
            // The element that this tooltip is attached to should not be included as an outside click
            filter(event => event.target !== this._elRef.nativeElement), map(_ => false)) : EMPTY;
        /**
         * Merge all open and close events into a single stream that emits a boolean that indicates whether
         * the tooltip should be opened or closed
         */
        const toggleEvents$ = merge(...openEvents$.concat(this._manualToggleEvent$, closeEvents$, outsideClick$));
        toggleEvents$
            .pipe(
        /**
         * This debounce prevents instantaneous opening and closing (or vice-versa) in the scenario where `openEvents` contains the
         * same event as `closeEvents`. For example, if both contain the "click" event, this will be fired twice in the space of a few ms
         */
        debounce(() => {
            if (this.openEvents.some(e => this.closeEvents.includes(e))) {
                return timer(5);
            }
            return of(0);
        }), filter(_ => !this.isDisabled), 
        /**
         * If this is an open event, use the input delay. Don't apply the delay to the close event
         */
        switchMap(isOpenEvent => {
            if (isOpenEvent && this.delay) {
                return of(isOpenEvent).pipe(delay(this.delay), takeUntil(this._cancelDelayedOpen$));
            }
            return of(isOpenEvent);
        }), takeUntil(this._destroyed$))
            .subscribe(isOpenEvent => {
            if (isOpenEvent) {
                const ref = this._open();
                /**
                 * No ref will be returned if the overlay is already attached
                 */
                if (ref) {
                    ref.changeDetectorRef.detectChanges();
                    ref.instance.close
                        .pipe(takeUntil(this._tooltipContainerDestroyed$))
                        .subscribe(_ => {
                        this.nwClose.emit();
                        this._close();
                    });
                    /**
                     * When the TooltipContainerComponent is destroyed we fire the _tooltipContainerDestroyed$
                     * so that our subscription to TooltipContainerComponent.close is unsubscribed from
                     */
                    ref.onDestroy(() => {
                        this._tooltipContainerDestroyed$.next();
                        this._tooltipContainerDestroyed$.complete();
                    });
                }
            }
            else {
                this._close();
            }
        });
    }
    _getPositionPair(placement) {
        // Include a 3px offset so that the tooltip is not flush with the element
        const offset = this._tooltipArrowSize + 3;
        const getXOffset = (placement) => {
            if (!this.withArrow) {
                return 0;
            }
            if (placement.startsWith('right')) {
                return offset;
            }
            if (placement.startsWith('left')) {
                return -offset;
            }
        };
        const getYOffset = (placement) => {
            if (!this.withArrow) {
                return 0;
            }
            if (placement.startsWith('top')) {
                return -offset;
            }
            if (placement.startsWith('bottom')) {
                return offset;
            }
        };
        const getPanelClass = (placement) => `tooltip-${placement}`;
        /**
         * The default position when no placement is specified
         */
        const bottom = new ConnectionPositionPair({ originX: "center", originY: "bottom" }, { overlayX: "center", overlayY: "top" }, null, getYOffset('bottom'), getPanelClass('bottom'));
        switch (placement) {
            case 'top':
                return new ConnectionPositionPair({ originX: "center", originY: "top" }, { overlayX: "center", overlayY: "bottom" }, null, getYOffset(placement), getPanelClass(placement));
            case 'top-start':
                return new ConnectionPositionPair({ originX: "start", originY: "top" }, { overlayX: "start", overlayY: "bottom" }, null, getYOffset(placement), getPanelClass(placement));
            case 'top-end':
                return new ConnectionPositionPair({ originX: "end", originY: "top" }, { overlayX: "end", overlayY: "bottom" }, null, getYOffset(placement), getPanelClass(placement));
            case 'bottom':
                return bottom;
            case 'bottom-start':
                return new ConnectionPositionPair({ originX: "start", originY: "bottom" }, { overlayX: "start", overlayY: "top" }, null, getYOffset(placement), getPanelClass(placement));
            case 'bottom-end':
                return new ConnectionPositionPair({ originX: "end", originY: "bottom" }, { overlayX: "end", overlayY: "top" }, null, getYOffset(placement), getPanelClass(placement));
            case 'right':
                return new ConnectionPositionPair({ originX: "end", originY: "center" }, { overlayX: "start", overlayY: "center" }, getXOffset(placement), null, getPanelClass(placement));
            case 'right-start':
                return new ConnectionPositionPair({ originX: "end", originY: "top" }, { overlayX: "start", overlayY: "top" }, getXOffset(placement), null, getPanelClass(placement));
            case 'right-end':
                return new ConnectionPositionPair({ originX: "end", originY: "bottom" }, { overlayX: "start", overlayY: "bottom" }, getXOffset(placement), null, getPanelClass(placement));
            case 'left':
                return new ConnectionPositionPair({ originX: "start", originY: "center" }, { overlayX: "end", overlayY: "center" }, getXOffset(placement), null, getPanelClass(placement));
            case 'left-start':
                return new ConnectionPositionPair({ originX: "start", originY: "top" }, { overlayX: "end", overlayY: "top" }, getXOffset(placement), null, getPanelClass(placement));
            case 'left-end':
                return new ConnectionPositionPair({ originX: "start", originY: "bottom" }, { overlayX: "end", overlayY: "bottom" }, getXOffset(placement), null, getPanelClass(placement));
            default:
                return bottom;
        }
    }
    _getPositionStrategy(placement) {
        /**
         * Format `placement` into a consistent data type of `Placement[]`
         */
        const placementsList = [placement].flat();
        /**
         * Get positions from preferred placements
         */
        const primaryPositions = placementsList.map(p => this._getPositionPair(p));
        /**
         * If `autoFlip` is enabled, include the inverse position of each `placement` input. Each of this inverse positions
         * will have a lower priority than each of the preferred positions generated from the `placement` input
         */
        const positions = this.autoFlip ?
            [...primaryPositions, ...placementsList.map(p => this._getPositionPair(placementFlipMap[p]))] :
            [...primaryPositions];
        return this._overlay
            .position()
            .flexibleConnectedTo(this.connectedTo || this._elRef)
            .withFlexibleDimensions(false)
            .withPositions(positions)
            .withPush(false);
    }
    _updatePositionStrategy(placement) {
        const positionStrategy = this._getPositionStrategy(placement);
        this._overlayRef.updatePositionStrategy(positionStrategy);
    }
    _getScrollStrategy(closeOnScroll) {
        if (closeOnScroll) {
            return this._overlay.scrollStrategies.close();
        }
        return this._overlay.scrollStrategies.reposition();
    }
    _updateScrollStrategy(closeOnScroll) {
        const scrollStrategy = this._getScrollStrategy(closeOnScroll);
        this._overlayRef.updateScrollStrategy(scrollStrategy);
    }
    ngOnDestroy() {
        this.hide();
        this._destroyed$.next();
        this._destroyed$.complete();
        this._overlayRef.dispose();
    }
}
TooltipDirective.ɵfac = function TooltipDirective_Factory(t) { return new (t || TooltipDirective)(i0.ɵɵdirectiveInject(i0.ElementRef), i0.ɵɵdirectiveInject(i1.Overlay), i0.ɵɵdirectiveInject(i0.ViewContainerRef), i0.ɵɵdirectiveInject(i0.Injector)); };
TooltipDirective.ɵdir = /*@__PURE__*/ i0.ɵɵdefineDirective({ type: TooltipDirective, selectors: [["", "nwTooltip", ""], ["", "nwPopover", ""]], inputs: { tooltip: [i0.ɵɵInputFlags.None, "nwTooltip", "tooltip"], popover: [i0.ɵɵInputFlags.None, "nwPopover", "popover"], context: "context", placement: "placement", isOpen: "isOpen", isDisabled: "isDisabled", delay: "delay", autoFlip: "autoFlip", openEvents: "openEvents", closeEvents: "closeEvents", containerClass: "containerClass", withArrow: "withArrow", withClose: "withClose", closeOnScroll: "closeOnScroll", closeOnOutsideClick: "closeOnOutsideClick", updatePositionOnAnimationFrame: "updatePositionOnAnimationFrame", connectedTo: "connectedTo", pointerEvents: "pointerEvents", breakpoint: "breakpoint", hostElementZIndex: "hostElementZIndex" }, outputs: { nwShown: "nwShown", nwHidden: "nwHidden", nwClose: "nwClose" }, exportAs: ["nw-tooltip", "nw-popover"], features: [i0.ɵɵNgOnChangesFeature] });
(() => { (typeof ngDevMode === "undefined" || ngDevMode) && i0.ɵsetClassMetadata(TooltipDirective, [{
        type: Directive,
        args: [{
                selector: '[nwTooltip],[nwPopover]',
                exportAs: 'nw-tooltip,nw-popover'
            }]
    }], () => [{ type: i0.ElementRef }, { type: i1.Overlay }, { type: i0.ViewContainerRef }, { type: i0.Injector }], { tooltip: [{
            type: Input,
            args: ['nwTooltip']
        }], popover: [{
            type: Input,
            args: ['nwPopover']
        }], context: [{
            type: Input
        }], placement: [{
            type: Input
        }], isOpen: [{
            type: Input
        }], isDisabled: [{
            type: Input
        }], delay: [{
            type: Input
        }], autoFlip: [{
            type: Input
        }], openEvents: [{
            type: Input
        }], closeEvents: [{
            type: Input
        }], containerClass: [{
            type: Input
        }], withArrow: [{
            type: Input
        }], withClose: [{
            type: Input
        }], closeOnScroll: [{
            type: Input
        }], closeOnOutsideClick: [{
            type: Input
        }], updatePositionOnAnimationFrame: [{
            type: Input
        }], connectedTo: [{
            type: Input
        }], pointerEvents: [{
            type: Input
        }], breakpoint: [{
            type: Input
        }], hostElementZIndex: [{
            type: Input
        }], nwShown: [{
            type: Output
        }], nwHidden: [{
            type: Output
        }], nwClose: [{
            type: Output
        }] }); })();
