import { ChangeDetectionStrategy, Component, EventEmitter, Inject, Input, Output } from "@angular/core";
import { DOCUMENT } from "@angular/common";
import { interval, of } from "rxjs";
import { catchError, filter, map, startWith, take, timeout } from "rxjs/operators";
import * as i0 from "@angular/core";
import * as i1 from "../resize-observer/resize-observer.directive";
const _c0 = ["*"];
export class WordCloudComponent {
    constructor(_elRef, _renderer, _cdRef, _document, elRef) {
        this._elRef = _elRef;
        this._renderer = _renderer;
        this._cdRef = _cdRef;
        this._document = _document;
        this.elRef = elRef;
        this.options = {};
        this.wordsPositioned = new EventEmitter();
        /**
         * Controls how many points are generated on the spiral. Should be kept between 0 and 1. Lower values generate more
         * points on the spiral but it can take a lot longer to position the words. If updating this value to <= 0.3, consider
         * all the places where this is used and for each one the amount of words in the cloud. It could also be a good idea to
         * make this value configurable by allowing it to be specified in the `options` input
         */
        this._spiralResolution = 0.3;
    }
    ngOnChanges(changes) {
        var _a, _b, _c, _d;
        const wordsChange = ((_a = changes.words) === null || _a === void 0 ? void 0 : _a.currentValue) !== ((_b = changes.words) === null || _b === void 0 ? void 0 : _b.previousValue);
        const optionsChange = ((_c = changes.options) === null || _c === void 0 ? void 0 : _c.currentValue) !== ((_d = changes.options) === null || _d === void 0 ? void 0 : _d.previousValue);
        if (wordsChange || optionsChange) {
            this._init();
        }
    }
    /**
     * Exports the canvas to a PNG image and returns the base64-encoded PNG data
     *
     * @returns A string containing the base64-encoded PNG data of the canvas
     */
    exportCanvas() {
        /**
         * Create a new canvas matching the dimensions of the original canvas. At this point, `_positionedWords` contains
         * the final positions of the words and we have no need to check for intersections, so we loop through our words
         * and draw them on our `exportCanvas`
         */
        const exportCanvas = this._renderer.createElement('canvas');
        const exportCtx = exportCanvas.getContext('2d');
        const ratio = window.devicePixelRatio;
        exportCanvas.width = this._canvas.width * ratio;
        exportCanvas.height = this._canvas.height * ratio;
        exportCtx.scale(ratio, ratio);
        this._positionedWords.forEach(pw => {
            const point = { x: pw.canvasX, y: pw.canvasY };
            this._setFontDetails(exportCtx, pw.fontSize);
            this._drawWord(pw, point, exportCtx);
        });
        /*
        * Get the lowest & highest words positions from the y-axis and based on that we get the word cloud content height
        * which we then use in order to extract only the content of the word cloud excluding white space and set it to a new canvas.
        */
        const lowestPointOnTheYAxis = Math.min(...this._positionedWords.map(pw => pw.canvasY - pw.height / 2));
        const highestPointOnTheYAxis = Math.max(...this._positionedWords.map(pw => pw.canvasY + pw.height / 2));
        const heightOfWordCloudContent = (highestPointOnTheYAxis - lowestPointOnTheYAxis) * ratio;
        const extractWordCloudImageData = exportCtx.getImageData(0, lowestPointOnTheYAxis * ratio, exportCanvas.width, heightOfWordCloudContent);
        const canvas = this._renderer.createElement('canvas');
        const context = canvas.getContext('2d');
        canvas.width = exportCanvas.width;
        canvas.height = heightOfWordCloudContent;
        context.putImageData(extractWordCloudImageData, 0, 0);
        return canvas.toDataURL('image/png');
    }
    downloadCanvas(filename) {
        const dataUrl = this.exportCanvas();
        const link = document.createElement('a');
        link.download = `${filename}.png`;
        link.href = dataUrl;
        link.click();
    }
    onResize() {
        this._init();
    }
    _init() {
        this.config = this._getConfig();
        this._truncatedWordsWithFontSize = this._getTruncatedWordsWithFontSize(this.words);
        this._positionedWords = [];
        this._drawCanvas();
        this._centerPoint = { x: this._canvas.width / 2, y: this._canvas.height / 2 };
        this._isFontLoaded().subscribe(() => {
            this._positionWords();
            if (this.config.debugMode) {
                console.info('Config', this.config);
                this._drawSprial();
            }
            this._cdRef.detectChanges();
        });
    }
    /**
     * Calculates the font size for each word based on its weight and the range of weights in the list of words
     *
     * @param words List of words with their respective weights
     * @returns A new list of words sorted by weight (largest to smallest) with their respective font sizes
     */
    _getTruncatedWordsWithFontSize(words) {
        const weights = words.map(w => w.weight);
        const minWeight = Math.min(...weights);
        const maxWeight = Math.max(...weights);
        return words.map(word => {
            return Object.assign(Object.assign({}, word), { fontSize: this._getFontSize(word.weight, minWeight, maxWeight), truncatedValue: this._truncateWord(word.value) });
        }).sort((a, b) => b.weight - a.weight);
    }
    _drawCanvas() {
        this._canvas = this._canvas || this._renderer.createElement('canvas');
        this._canvas.width = this._elRef.nativeElement.parentElement.clientWidth;
        this._canvas.height = this._elRef.nativeElement.parentElement.clientHeight;
        this._ctx = this._canvas.getContext('2d');
        if (this.config.debugMode) {
            this._renderer.appendChild(this._elRef.nativeElement, this._canvas);
        }
    }
    /**
     * This method is responsible for positioning the words onto a canvas by iterating through each word, finding a suitable location for
     * it using a spiral algorithm, and checking if the word's bounding box intersects with any existing bounding boxes
     */
    _positionWords() {
        const t1 = performance.now();
        /**
         * Sets the font details, places the word on the spiral, and calculates the bounding box of the word based on its font size. If the
         * bounding box intersects with another bounding box, the index is incremented and we try to place it again
         */
        const positionWord = (wordWithFontSize, index = 0) => {
            this._setFontDetails(this._ctx, wordWithFontSize.fontSize);
            const point = this._placeOnSpiral(index);
            const metrics = this._ctx.measureText(wordWithFontSize.truncatedValue);
            const fontHeight = (metrics.actualBoundingBoxAscent + metrics.actualBoundingBoxDescent) * 1.1;
            const { paddingX, paddingY } = this.config;
            const width = metrics.width + paddingX;
            const height = fontHeight + paddingY;
            /**
             * Adjust the x and y based on textAlign = "center" and textBaseline = "middle";
             */
            const boundingBox = {
                x: point.x - (width / 2),
                y: point.y - (height / 2),
                width,
                height
            };
            /**
             * If this bounding box intersects another bounding box, increment the index and try the next place
             */
            if (this._isIntersecting(boundingBox)) {
                return positionWord(wordWithFontSize, index + 1);
            }
            this._positionedWords = this._positionedWords.concat(Object.assign(Object.assign(Object.assign({}, boundingBox), wordWithFontSize), { canvasX: point.x, canvasY: point.y }));
            if (this.config.debugMode) {
                this._ctx.strokeStyle = 'blue';
                this._ctx.strokeRect(boundingBox.x, boundingBox.y, boundingBox.width, boundingBox.height);
                this._drawWord(wordWithFontSize, point, this._ctx);
            }
        };
        this._truncatedWordsWithFontSize.forEach(word => {
            positionWord(word);
        });
        this._positionedWords = this._scalePositionedWords(this._positionedWords);
        /**
         * Sort words before emitting so that a trackBy function will work correctly with the collection
         */
        const sortedWords = this._positionedWords.sort((a, b) => a.value.localeCompare(b.value));
        this.wordsPositioned.emit(sortedWords);
        if (this.config.debugMode) {
            console.info(`Words positioned on sprial in ${performance.now() - t1}ms`);
        }
    }
    _truncateWord(value) {
        if (value.length > this.config.maxCharCount) {
            return value.substring(0, this.config.maxCharCount) + '...';
        }
        return value;
    }
    _drawWord(wordWithFontSize, point, ctx) {
        ctx.fillStyle = wordWithFontSize.exportColor;
        ctx.fillText(wordWithFontSize.truncatedValue, point.x, point.y);
    }
    _setFontDetails(ctx, fontSize) {
        ctx.font = `${this.config.fontWeight} ${fontSize}px/1 ${this.config.fontFamily}`;
        ctx.textAlign = "center";
        ctx.textBaseline = "middle";
    }
    _isIntersecting(boundingBox) {
        /**
         * Check whether or not two bounding boxes overlap
         */
        const doBoxesOverlap = (boxA, boxB) => {
            return !(boxA.x + boxA.width < boxB.x ||
                boxB.x + boxB.width < boxA.x ||
                boxA.y + boxA.height < boxB.y ||
                boxB.y + boxB.height < boxA.y);
        };
        /**
         * Check if any bounding boxes overlap
         */
        return this._positionedWords.some(word => doBoxesOverlap(boundingBox, word));
    }
    /**
     * Calculate the font size for a given word weight based on the minimum and maximum weights and the minimum and maximum font sizes
     * defined in the word cloud configuration object
     */
    _getFontSize(wordWeight, minWeight, maxWeight) {
        const weightDiff = maxWeight - minWeight;
        const factor = (wordWeight - minWeight) / weightDiff;
        const fontSize = ((this.config.maxFontSize - this.config.minFontSize) * factor) + this.config.minFontSize;
        /**
         * fontSize may be NaN if all words have the same weight. If they do, return the fontSize that is midway
         * between the minFontSize and the maxFontSize
         */
        if (fontSize) {
            return Math.round(fontSize);
        }
        return Math.round((this.config.minFontSize + this.config.maxFontSize) / 2);
    }
    /**
     * @description This method is used to place a point on a spiral based on the input n. It calculates the coordinates of the
     * point on the spiral using the spiral helper function and returns an object of type IPoint. The spiral is defined using a combination
     * of trigonometric functions and aspects such as factor, aspectRatio, and angle.
     *
     * @returns An object of type `IPoint` with `x` and `y` properties
     */
    _placeOnSpiral(n) {
        const spiral = (i) => {
            const aspectRatio = this._canvas.width / this._canvas.height;
            const angle = i;
            const x = this._centerPoint.x + (1 + angle) * Math.cos(angle) * this._spiralResolution * aspectRatio;
            const y = this._centerPoint.y + (1 + angle) * Math.sin(angle) * this._spiralResolution;
            return { x, y };
        };
        return spiral(n);
    }
    _isOutsideCanvas(boundingBox, canvasWidth, canvasHeight) {
        return (boundingBox.x + boundingBox.width) > canvasWidth ||
            boundingBox.x < 0 ||
            (boundingBox.y + boundingBox.height) > canvasHeight ||
            boundingBox.y < 0;
    }
    /**
     * Retrieve the word cloud configuration object by merging default values with any options passed to this component instance
     */
    _getConfig() {
        const defaultConfig = {
            debugMode: false,
            fontFamily: 'ProximaNova',
            fontWeight: 'normal',
            maxFontSize: 48,
            minFontSize: 16,
            paddingX: 8,
            paddingY: 8,
            maxCharCount: 20,
            resizeTolerance: 0
        };
        return Object.assign(Object.assign({}, defaultConfig), this.options);
    }
    _drawSprial() {
        this._ctx.save();
        this._ctx.fillStyle = 'white';
        for (let i = 0; i < 3000; i++) {
            const { x, y } = this._placeOnSpiral(i);
            this._ctx.fillRect(x, y, 1, 1);
        }
        this._ctx.restore();
    }
    /**
     * Determine the amount of overflow for the words that have been placed outside the canvas, calculate the scaling factor for repositioning these
     * words back inside the canvas, and update and return the placed words with the scaling factor applied
     */
    _scalePositionedWords(positionedWords) {
        const outOfBounds = positionedWords.filter(word => this._isOutsideCanvas(word, this._canvas.width, this._canvas.height));
        /**
         * Determine the minimum and maximum x-coordinates of the words that are outside the canvas. The minimum x-coordinate is the leftmost
         * point of any out-of-bounds word, and the maximum x-coordinate is the rightmost point of any out-of-bounds word
         */
        const minX = Math.min(...outOfBounds.map(oob => oob.x));
        const maxX = Math.max(...outOfBounds.map(oob => oob.x + oob.width));
        /**
         * Calculate the amount of overflow for the words that are outside the canvas. `overflowLeft` is the amount by which the out-of-bounds words
         * exceed the left edge of the canvas, and `overflowRight` is the amount by which they exceed the right edge of the canvas
         */
        const overflowLeft = minX < 0 ? Math.abs(minX) : 0;
        const overflowRight = maxX > this._canvas.width ? (maxX - this._canvas.width) : 0;
        /**
         * Find the largest horizontal overflow and double it. As we're scaling towards the center, rather than summing the left
         * and right overflows we need to take the largest and double it to ensure we don't end up with words being positioned out
         * of bounds on the x-axis
         */
        const overflowX = Math.max(overflowLeft, overflowRight) * 2;
        const xScale = this._canvas.width / (this._canvas.width + overflowX);
        if (this.config.debugMode) {
            console.info(`minX: ${minX}, maxX: ${maxX}, overflowX: ${overflowX}, xScale: ${xScale}`);
        }
        /**
         * Determine the minimum and maximum y-coordinates of the words that are outside the canvas. The minimum y-coordinate is the topmost
         * point of any out-of-bounds word, and the maximum Y-coordinate is the bottom-most point of any out-of-bounds word
         */
        const minY = Math.min(...outOfBounds.map(oob => oob.y));
        const maxY = Math.max(...outOfBounds.map(oob => oob.y + oob.height));
        const overflowTop = minY < 0 ? Math.abs(minY) : 0;
        const overflowBottom = maxY > this._canvas.height ? (maxY - this._canvas.height) : 0;
        /**
         * Find the largest vertical overflow and double it. As we're scaling towards the center, rather than summing the top
         * and bottom overflows we need to take the largest and double it to ensure we don't end up with words being positioned out
         * of bounds on the y-axis
         */
        const overflowY = Math.max(overflowTop, overflowBottom) * 2;
        const yScale = this._canvas.height / (this._canvas.height + overflowY);
        const minScale = Math.min(xScale, yScale);
        if (this.config.debugMode) {
            console.info(`minY: ${minY}, maxY: ${maxY}, overflowY: ${overflowY}, yScale: ${yScale}`);
        }
        const moveTowardsCenter = (x, y, height, scale) => {
            const scaledHeightOffset = (height / 4) * scale;
            const dx = this._centerPoint.x - x;
            const dy = this._centerPoint.y - y + scaledHeightOffset;
            /**
             * Determine the distance between the current position and the target position (the center) and
             * multiply it by our scale (between 0 and 1) to move partway towards the target
             */
            const dist = Math.sqrt((dx * dx) + (dy * dy)) * (1 - scale);
            /**
             * Calculate the angle at which we should move and the update our x and y using our calculated distance and angle
             *
             * ref: https://stackoverflow.com/a/5995931/1128290
             */
            const angle = Math.atan2(dy, dx);
            return {
                x: x + (dist * Math.cos(angle)),
                y: y + (dist * Math.sin(angle))
            };
        };
        return this._positionedWords.map(pw => {
            const { x, y } = moveTowardsCenter(pw.x, pw.y, pw.height, minScale);
            const { x: canvasX, y: canvasY } = moveTowardsCenter(pw.canvasX, pw.canvasY, pw.height, minScale);
            const width = pw.width * minScale;
            const height = pw.height * minScale;
            const fontSize = pw.fontSize * minScale;
            return Object.assign(Object.assign({}, pw), { canvasX,
                canvasY,
                x,
                y,
                width,
                height,
                fontSize });
        });
    }
    _isFontLoaded() {
        const fontTimeout = 5000;
        /**
         * Needs to be a valid font string
         * ref: https://developer.mozilla.org/en-US/docs/Web/CSS/font
         */
        const fontString = `${this.config.fontWeight} 12px ${this.config.fontFamily}`;
        return interval(250).pipe(startWith(0), map(_x => this._document.fonts.check(fontString)), filter(isLoaded => isLoaded), map(isLoaded => ({ isLoaded })), take(1), timeout(fontTimeout), catchError(_e => {
            console.warn(`The font "${fontString}" failed to load within ${fontTimeout}ms`);
            return of({ isLoaded: false });
        }));
    }
}
WordCloudComponent.ɵfac = function WordCloudComponent_Factory(t) { return new (t || WordCloudComponent)(i0.ɵɵdirectiveInject(i0.ElementRef), i0.ɵɵdirectiveInject(i0.Renderer2), i0.ɵɵdirectiveInject(i0.ChangeDetectorRef), i0.ɵɵdirectiveInject(DOCUMENT), i0.ɵɵdirectiveInject(i0.ElementRef)); };
WordCloudComponent.ɵcmp = /*@__PURE__*/ i0.ɵɵdefineComponent({ type: WordCloudComponent, selectors: [["nw-word-cloud"]], inputs: { words: "words", options: "options" }, outputs: { wordsPositioned: "wordsPositioned" }, features: [i0.ɵɵNgOnChangesFeature], ngContentSelectors: _c0, decls: 4, vars: 6, consts: [["nwResizeObserver", "", 1, "nw-word-cloud-container", 3, "nwResize", "element", "tolerance"], [1, "font-loader"]], template: function WordCloudComponent_Template(rf, ctx) { if (rf & 1) {
        i0.ɵɵprojectionDef();
        i0.ɵɵelementStart(0, "div", 0);
        i0.ɵɵlistener("nwResize", function WordCloudComponent_Template_div_nwResize_0_listener() { return ctx.onResize(); });
        i0.ɵɵelementStart(1, "div", 1);
        i0.ɵɵtext(2, " This element is hidden and is used to delay drawing to the canvas until the font is loaded. If we don't wait and the font has not loaded, the word will be drawn to the canvas with a fallback font and, as a result, will not be positioned correctly when used with the supplied font details ");
        i0.ɵɵelementEnd();
        i0.ɵɵprojection(3);
        i0.ɵɵelementEnd();
    } if (rf & 2) {
        i0.ɵɵproperty("element", ctx.elRef.nativeElement)("tolerance", ctx.config.resizeTolerance);
        i0.ɵɵadvance();
        i0.ɵɵstyleProp("font-family", ctx.config.fontFamily)("font-weight", ctx.config.fontWeight);
    } }, dependencies: [i1.ResizeObserverDirective], encapsulation: 2, changeDetection: 0 });
(() => { (typeof ngDevMode === "undefined" || ngDevMode) && i0.ɵsetClassMetadata(WordCloudComponent, [{
        type: Component,
        args: [{ selector: 'nw-word-cloud', changeDetection: ChangeDetectionStrategy.OnPush, template: "<div class=\"nw-word-cloud-container\"\n    nwResizeObserver\n    [element]=\"elRef.nativeElement\"\n    [tolerance]=\"config.resizeTolerance\"\n    (nwResize)=\"onResize()\">\n    <div class=\"font-loader\"\n        [style.fontFamily]=\"config.fontFamily\"\n        [style.fontWeight]=\"config.fontWeight\">\n        This element is hidden and is used to delay drawing to the canvas until the font is loaded. If we don't wait and the\n        font has not loaded, the word will be drawn to the canvas with a fallback font and, as a result, will not be positioned\n        correctly when used with the supplied font details\n    </div>\n    <ng-content></ng-content>\n</div>" }]
    }], () => [{ type: i0.ElementRef }, { type: i0.Renderer2 }, { type: i0.ChangeDetectorRef }, { type: Document, decorators: [{
                type: Inject,
                args: [DOCUMENT]
            }] }, { type: i0.ElementRef }], { words: [{
            type: Input
        }], options: [{
            type: Input
        }], wordsPositioned: [{
            type: Output
        }] }); })();
(() => { (typeof ngDevMode === "undefined" || ngDevMode) && i0.ɵsetClassDebugInfo(WordCloudComponent, { className: "WordCloudComponent" }); })();
